/*
 * fbshot.c -- FrameBuffer Screen Capture Utility
 * (C)opyright 2000 U4X Labs
 *
 * Originally Written by: Stephan Beyer <PH-Linex@gmx.net>
 * Maintained by: Paul Mundt <lethal@u4x.org>
 * 
 * 	This is a simple program that generates a
 * screenshot of the specified framebuffer device and
 * terminal and writes it to a specified file using
 * the PNG format.
 *
 * See ChangeLog for modifications, CREDITS for credits.
 * 
 * fbshot is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either Version 2
 * of the License, or (at your option) any later version.
 *
 * fbshot is distributed in the hope that it will be useful, but
 * WITHOUT ANY  WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License with fbshot; if not, please write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 
 * 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <mntent.h>
#include <linux/fb.h>
#include <sys/vt.h>
#include <png.h>

#define PACKAGE "fbshot"
#define	VERSION	"0.2"
#define	DEFAULT_FB "/dev/fb0"
#define MAINTAINER_NAME "Paul Mundt"
#define MAINTAINER_ADDR "lethal@chaoticdreams.org"

#define display_usage() fprintf(stderr, "Usage: %s [-ghi] [-{C|c} vt] [-d dev] [-s n] filename.png\n", PACKAGE); 

/* some conversion macros */
#define RED565(x) ((((x) >> (5+6)) & 0x1f) << 3)
#define GREEN565(x) ((((x) >> 6) & 0x3f) << 2)
#define BLUE565(x) (((x) & 0x1f) << 3)

#define ALPHA1555(x) (((x) >> (5+5+5)) & 0x1)
#define RED1555(x) ((((x) >> (5+5)) & 0x1f) << 3)
#define GREEN1555(x) ((((x) >> 5) & 0x1f) << 3)
#define BLUE1555(x) (((x) & 0x1f) << 3)

static struct fb_fix_screeninfo fb_fixinfo;
static struct fb_var_screeninfo fb_varinfo;
static struct fb_cmap fb_cmapinfo;
static int interlace=PNG_INTERLACE_ADAM7;
static int gray=0; /* -1 on ; 0 off ; */
static int waitbfg=0; /* wait before grabbing (for -C )... */

static void display_help()
{
    fprintf(stdout,"FBShot, version %s\n"
		   "(C)opyright 2000 U4X Labs\n\n",
		   VERSION);
		   
    display_usage();
    
    fprintf(stdout,"\nOptions:\n"
		   "\t--slowcon n,  -C n  \tgrab from console n, for slower framebuffers\n"
		   "\t--console n,  -c n  \tgrab from console n\n"
		   "\t--device dev, -d dev\tuse framebuffer device dev instead of %s\n"
		   "\t--grascale,   -g    \tsave PNG in grayscale\n"
		   "\t--noint,      -i    \tturns off interlacing\n"
		   "\t--sleep n,    -s n  \tsleep for n seconds before making screenshot\n"
		   "\t--help,       -h    \tprint this text\n\n"
		   "Report bugs to %s <%s>\n",
		   DEFAULT_FB, MAINTAINER_NAME, MAINTAINER_ADDR);
}


static void chvt(int num)
 {
	int fd;

	if(!(fd = open("/dev/console", O_RDWR)))
	 {
		perror("open");
		exit(1);
	 }

	if (ioctl(fd, VT_ACTIVATE, num))
	 {
		perror("ioctl VT_ACTIVATE");
		exit(1);
	 }

	if (ioctl(fd, VT_WAITACTIVE, num))
	 {
		perror("ioctl VT_WAITACTIVE");
		exit(1);
	 }

	close(fd);

	if (waitbfg)
	    sleep (3);
 }


static char *read_framebuffer(char *device, int vt_num)
 {
    int fd, old_vt, i;
    struct vt_stat vt_info;
    char buffer[7680000]; /* is there a more elegant solution? eg with
			    a pointer char *buffer? */

    /* first get some information about the _current_ console */
    if (vt_num!=-1)
     {
	char tmp[128];
	if ((fd = open("/dev/console", O_RDONLY)) == -1) 
	 {
	    perror("Could not open /dev/console");
	    exit(1);
	 }
	if (ioctl(fd, VT_GETSTATE, &vt_info))
	 {
	    perror ("ioctl VT_GETSTATE");
	    exit(1);
	 }
	close (fd);
	old_vt=vt_info.v_active;
     }

    /* now open framebuffer device */
    if(!(fd=open(device, O_RDONLY)))
     {
	fprintf (stderr, "Error: Couldn't open %s.\n", device);
	exit(1);
     }

    if (vt_num!=-1)
     {
	chvt(vt_num); /* go there for information */
	if (waitbfg)
	    sleep (1);
     }

    
    if (ioctl(fd, FBIOGET_FSCREENINFO, &fb_fixinfo))
     {
	perror ("ioctl FBIOGET_FSCREENINFO");
	exit(1);
     } /* get fixed info */
    if (ioctl(fd, FBIOGET_VSCREENINFO, &fb_varinfo))
     {
	perror ("ioctl FBIOGET_VSCREENINFO");
	exit(1);
     } /* get variable info */
#if 0
    if (ioctl(fd, FBIOGETCMAP, &fb_cmapinfo))
     {
	perror ("ioctl FBIOGETCMAP");
	exit(1);
     } /* get cmap */
#endif


    if (vt_num!=-1)
	chvt(vt_info.v_active); /* go back for printing */

    fprintf(stdout, "Framebuffer %s is %i bytes.\n", device, 
	            (fb_varinfo.xres * fb_varinfo.yres * 
		    (fb_varinfo.bits_per_pixel >> 3)));

    fprintf(stdout, "Grabbing %ix%i ... ", fb_varinfo.xres, fb_varinfo.yres);
    fflush(stdout);

    if (vt_num!=-1)
     {
	chvt(vt_num); /* go there for reading */
	if (waitbfg)
	    sleep (3);
     }
    read(fd, buffer, (fb_varinfo.xres * fb_varinfo.yres * (fb_varinfo.bits_per_pixel >> 3))); 
	/* read full framebuffer image data */

    if (vt_num!=-1)
	chvt(vt_info.v_active); /* go back for printing */

    fprintf(stdout,"done.\n"); 
    close (fd);

    return buffer;
 }


static char *convert8to24(int width, int height, char *buffer)
 {
    int i;
    char nbuffer[7680000];
    
    for (i=0; i<height*width; i++)
     {
        unsigned char c=buffer[i];
        unsigned char outc[3];
        unsigned short r[255], g[255], b[255];

	outc[0]=r[c]>>8;
	outc[1]=g[c]>>8;
	outc[2]=b[c]>>8;
	outc[3]=0>>8;
	
	sprintf(nbuffer, "%s%s", nbuffer, outc);
     }
    
    return strdup(nbuffer);
 }

static char *convert1555to32(int width, int height, char *buffer)
 {
    int i;
    char nbuffer[7680000];
    
    for (i=0; i<height*width; i++)
     {
        __u16 *cb=(__u16 *) &buffer[i];
        unsigned char outc[3];

	outc[0]=(char)RED1555(*cb);
	outc[1]=(char)GREEN1555(*cb);
	outc[2]=(char)BLUE1555(*cb);
	outc[3]=(char)0; 
	/* (char)ALPHA1555(*cb) may be the same, because alpha should 
	    be 0 */
	
	sprintf(nbuffer, "%s%s", nbuffer, outc);
     }
    
    return strdup(nbuffer);
 }

static char *convert565to32(int width, int height, char *buffer)
 {
    int i;
    char nbuffer[7680000];
    
    for (i=0; i<height*width; i++)
     {
        __u16 *cb=(__u16 *) &buffer[i];
        unsigned char outc[3];

	outc[0]=(char)RED565(*cb);
	outc[1]=(char)GREEN565(*cb);
	outc[2]=(char)BLUE565(*cb);
	outc[3]=(char)0;
	
	sprintf(nbuffer, "%s%s", nbuffer, outc);
     }
    
    return strdup(nbuffer);
 }


static int Write_PNG(char *buffer, char *filename, int width, int height, int bits)
 {
    png_bytep row_pointers[height];

    png_structp png_ptr;
    png_infop info_ptr;
    int i;
    int bit_depth=0, color_type; 

    FILE *OUTfd = fopen(filename, "wb");

    for (i=0; i<height; i++)
	row_pointers[i]=buffer+i*4*width;

    fprintf(stdout, "Writing to %s ... ", filename);
    fflush(stdout);

    if (!OUTfd)
     {
	fprintf (stderr, "Error: Couldn't fopen %s.\n", filename);
	exit(1);
     }
    
    png_ptr = png_create_write_struct(
	PNG_LIBPNG_VER_STRING, 
	(png_voidp)NULL, (png_error_ptr)NULL, (png_error_ptr)NULL);
	
    if (!png_ptr)
     {
	fprintf (stderr, "Error: Couldn't create PNG write struct.\n");
	exit(1);
     }

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
     {
	png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
	fprintf (stderr, "Error: Couldn't create PNG info struct.\n");
	exit(1);
     }
    
    png_init_io(png_ptr, OUTfd);
    
    png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
    
    /* here please add some things, please, please, please */

    switch(bits) {
	case 8:
/*	     buffer=convert8to24(width, height, buffer); */

	    /* ... I do a break instead... */
	    break;
	case 24: /* NOTHING TESTED */
	    if (gray)
	     {
	     	bit_depth=8;
		color_type=PNG_COLOR_TYPE_GRAY;
	     } else {
		bit_depth=4;
		color_type=PNG_COLOR_TYPE_RGB;
	     }
	    png_set_invert_alpha(png_ptr);
	    png_set_bgr(png_ptr);

	    fprintf(stdout, "Note: This mode is currently untested, send results to %s\n", MAINTAINER_ADDR);
	    break;
	case 15:
	    /* 5/6/5 */
/*	    buffer=convert565to32(width, height, buffer);*/

	    /* and then the same as 32 bit */
	    break;
	case 16:
    	    /* 1/5/5/5 */
/*	    buffer=convert1555to32(width, height, buffer);*/
	    
	    /* and then the same as 32 bit */

	    break;
	case 32:
	    if (gray)
	     {
		bit_depth=16; /* not as original as converting would be, */
		color_type=PNG_COLOR_TYPE_GRAY_ALPHA; /* but it looks good. */
	     } else {
		bit_depth=8;
		color_type=PNG_COLOR_TYPE_RGB_ALPHA;
	     }

	    png_set_invert_alpha(png_ptr);
	    png_set_bgr(png_ptr);

	    break;
     }
    
    if (bit_depth==0)
     {
	fprintf (stderr, "%d bits per pixel are not yet supported! ", bits);
	fprintf (stderr, "But you may write it...\n");
	exit(1);
     }

    png_set_IHDR(png_ptr, info_ptr, width, height, 
	bit_depth, color_type, interlace, 
	PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    
    png_write_info(png_ptr, info_ptr);

    png_write_image(png_ptr, row_pointers);

    png_write_end(png_ptr, info_ptr);
    /* puh, done, now freeing memory... */
    png_destroy_write_struct(&png_ptr, &info_ptr);

    fclose(OUTfd);
    return 0;
 }

static char optstring[] = "ghiC:c:d:s:";
static struct option long_options[] = {
	{"slowcon", 1, 0, 'C'},
	{"console", 1, 0, 'c'},
	{"device", 1, 0, 'd'},
	{"grayscale", 0, 0, 'g'},
	{"help", 0, 0, 'h'},
	{"noint", 0, 0, 'i'},
	{"sleep", 1, 0, 's'},
	{0, 0, 0, 0}
};

int main(int argc, char **argv)
 {
    char *buffer, *device = NULL, *outfile = argv[argc-1];
    int optc, vt_num=-1;

    for(;;)
     {
	int optind = 0;

	if ((optc = getopt_long(argc, argv, optstring,
	                        long_options, &optind)) == -1)
		break;

	switch (optc) {
/* please keep this list alphabetical */ 
	case 'C':
	    waitbfg=1;
	case 'c':
	    vt_num=atoi(optarg);
	    break;
	case 'd':
	    device=optarg;
	    break;
	case 'g':
	    gray=-1;
	    break;
	case 'h':
	    display_help();
	    return 1;
	    break;
	case 'i':
	    interlace=PNG_INTERLACE_NONE;
	    break;
	case 's':
	    sleep (atoi(optarg));
	    break;
	default:
	    break;
	}
     }
    
    if ((optind==argc) ||
	(1!=argc-optind))
     {
	display_usage();
	return 1;
     }
    
    if (NULL==device)
     {
        device=getenv("FRAMEBUFFER");
	if (NULL==device)
	 {
	    device=DEFAULT_FB;
	 }
     }

    buffer = read_framebuffer(device, vt_num);

    if (buffer == NULL) {
	fprintf(stderr, "An error occured.\n");
	return 1;
    }

    Write_PNG(buffer, outfile, fb_varinfo.xres, fb_varinfo.yres, 
					fb_varinfo.bits_per_pixel);
    
    printf ("done.\n");
    
    return 0;
 }

